#include <stdlib.h>

#include "vtkRenderer.h"
#include "vtkRenderWindow.h"
#include "vtkActor.h"
#include "vtkFollower.h"
#include "vtkActorCollection.h"
#include "vtkPolyDataMapper.h"
#include "vtkLineSource.h"
#include "vtkTransform.h"
#include "vtkSphereSource.h"
#include "vtkAxes.h"

#include "defs.h"
#include "types.h"
#include "protos.h"


extern vtkPolyDataMapper **sphereMappers;
extern vtkPolyDataMapper *bondMappers[2];
extern vtkPolyDataMapper *lineMapper;
extern vtkPolyDataMapper *axesMapper;
extern vtkPolyDataMapper **labelMappers;
extern vtkPolyDataMapper *surfaceMapper;
extern vtkRenderWindow *renWindow;

extern SURFACE_INFO surface_info;

// Draw the specified frame.  Drawing consists of creating the necessary
// actors and adding them to the vtk renderer.  Lastly, the render window
// is updated to reflect the changes.
void draw_frame(vtkRenderer *ren, vtkRenderWindow *renWindow, DRAW_FLAGS *flags,
	   FRAME_LIST *frame, ELEMENT_LIST *element_head){

  if(frame == NULL) return;

  // First delete the contents of the current frame
  delete_frame(ren);
  
  // Initialize actors for the current frame
  init_actors(ren, flags, frame, element_head);

  // Render the scene in the VTK window
  renWindow->Render();
}

// Draw all the atoms in the specified frame.  Each atom's position
// and color is dependent on the atom's element, which is linked to
// in the ATOM_LIST structure.
void draw_atoms(vtkRenderer *ren, DRAW_FLAGS *flags, FRAME_LIST *frame){
  ATOM_LIST *atoms;

  atoms = frame->atoms;

  while(atoms != NULL){
    vtkActor *sphereActor = vtkActor::New();
    sphereActor->SetMapper(sphereMappers[atoms->info->sphere_index]);
    sphereActor->GetProperty()->SetColor(atoms->info->color.red,
					 atoms->info->color.green,
					 atoms->info->color.blue);
    sphereActor->SetPosition(atoms->x, atoms->y, atoms->z);
    ren->AddActor(sphereActor);
    sphereActor->Delete();
    
    atoms = atoms->next;
  }
}

// Draw the bonds for the specified frame.  The drawing flags specify
// what type of bond to draw.
void draw_bonds(vtkRenderer *ren, DRAW_FLAGS *flags, FRAME_LIST *frame,
		ELEMENT_LIST *element_head){
  
  BOND_LIST *bonds;
  ELEMENT_LIST *elements;
  COLOR bond_color;
  int bond_complex = flags->bond_complex;
  vtkTransform *trans;

  if(frame == NULL) return;

  bonds = frame->bonds;

  // find the element which stores the bond color
  elements = element_head;
  while( strncmp(elements->label, BONDC_LABEL, LABEL_LEN) != 0 && elements != NULL){
    elements = elements->next;
  } 
  
  if(elements == NULL){ // If it didn't find one then give it a default color
    bond_color = flags->bond_color;
  }
  else{
    bond_color = elements->color;
  }

  // Create the actors for each of the bonds
  if(bond_complex == LINE_BOND){
    
    while(bonds != NULL){
	
      vtkActor *bondActor = vtkActor::New();
      bondActor->SetMapper(lineMapper);
      bondActor->GetProperty()->SetLineWidth(3.0);
      bondActor->GetProperty()->SetColor(bond_color.red, bond_color.green, bond_color.blue);
      trans = getTransformation(bonds->end1->x, bonds->end1->y, bonds->end1->z,
				bonds->end2->x, bonds->end2->y, bonds->end2->z);
      bondActor->SetUserMatrix(trans->GetMatrix());
      ren->AddActor(bondActor);
      trans->Delete();
      bondActor->Delete();
      
      bonds = bonds->next;

      
    }
    
  }
  else if(bond_complex == STRC_BOND){
    
    while(bonds != NULL){
      vtkActor *bondActor = vtkActor::New();
      bondActor->SetMapper(bondMappers[0]);
      bondActor->GetProperty()->SetColor(bond_color.red, bond_color.green, bond_color.blue);
      trans = getTransformation(bonds->end1->x, bonds->end1->y, bonds->end1->z,
				bonds->end2->x, bonds->end2->y, bonds->end2->z);

      // vtkCylinderSource cylinders are aligned along the y-axis and centered at the
      // origin, so they take extra transformations in order to use the getTransformation
      // function I wrote for the line bonds(which are aligned along the z-axis and start
      // at the origin.
      trans->RotateX(90.0);
      trans->Translate(0.0, 0.5, 0.0);

      bondActor->SetUserMatrix(trans->GetMatrix());
      ren->AddActor(bondActor);
      trans->Delete();
      bondActor->Delete();
      
      bonds = bonds->next;
    }
  }
  else if(bond_complex == TAPC_BOND){
    while(bonds != NULL){
      vtkActor *bondActor = vtkActor::New();
      bondActor->SetMapper(bondMappers[1]);
      bondActor->GetProperty()->SetColor(bond_color.red, bond_color.green, bond_color.blue);
      trans = getTransformation(bonds->end1->x, bonds->end1->y, bonds->end1->z,
				bonds->end2->x, bonds->end2->y, bonds->end2->z);
      
      // vtkCylinderSource cylinders are aligned along the y-axis and centered at the
      // origin, so they take extra transformations in order to use the getTransformation
      // function I wrote for the line bonds(which are aligned along the z-axis and start
      // at the origin.
      trans->RotateX(90.0);
      trans->Translate(0.0, 0.5, 0.0);
      bondActor->SetUserMatrix(trans->GetMatrix());
      ren->AddActor(bondActor);
      trans->Delete();
      bondActor->Delete();

      bonds = bonds->next;
    }
  }

  
}

// Draw the coordinate axes.
void draw_axes(vtkRenderer *ren){

  vtkActor *axesActor = vtkActor::New();
  axesActor->SetMapper(axesMapper);
  axesActor->GetProperty()->SetLineWidth(3.0);

  ren->AddActor(axesActor);

  axesActor->Delete();

}

// Draw the atom(element) labels.  vtkFollower is a subclass of
// vtkActor which always faces the camera.
void draw_atom_labels(vtkRenderer *ren, DRAW_FLAGS *flags, ATOM_LIST *atoms){
  float radius = flags->sphr_size;

  while(atoms != NULL){
    vtkFollower *labelActor = vtkFollower::New();
    labelActor->SetMapper(labelMappers[atoms->info->sphere_index]);
    labelActor->SetPosition(atoms->x, atoms->y + (atoms->info->radius * radius), atoms->z);
    labelActor->SetScale(0.1, 0.1, 0.1);
    labelActor->SetCamera(ren->GetActiveCamera());
    ren->AddActor(labelActor);
    labelActor->Delete();

    atoms = atoms->next;
  }

}

// Draw isosurfaces.
void draw_isosurfaces(vtkRenderer *ren){

  vtkActor *surfaceActor = vtkActor::New();
  surfaceActor->SetMapper(surfaceMapper);
  surfaceActor->GetProperty()->SetSpecular(surface_info.specular);
  surfaceActor->GetProperty()->SetSpecularPower(surface_info.spec_power);
  
  // The position marks the location of the first sample from the density dataset.
  // I'm assuming that the dataset exists entirely within the 1st octect(+ x,y,z).
  surfaceActor->SetPosition(0.0,0.0,0.0);

  ren->AddActor(surfaceActor);
  
  surfaceActor->Delete();
}
